/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */

package org.w3c.rdf.model.impl;

import org.w3c.rdf.model.*;
import org.w3c.rdf.util.RDFDigestUtil;
import org.w3c.tools.crypt.*;

import java.util.*;

/**
 * @author Sergey Melnik <melnik@db.stanford.edu>
 */

public class ModelImpl implements Model, Digestable, Digest {

  //  static LongTable resources = new LongTable(); // long -> ResourceImpl
  //  static LongTable literals = new LongTable(); // long -> LiteralImpl

  //  static Hashtable resources = new Hashtable();
  //  static Hashtable literals = new Hashtable();

  // whether triples and lookup are shared
  boolean shared = false;
  byte[] digest;

  //  int findCounter = 0;

  String uri;
  Hashtable triples;
  //  LongTable subjects, predicates, objects;

  // first find operation creates lookup table
  LongTable lookup;

  public ModelImpl() {
    triples = new Hashtable();
    lookup = new LongTable(true);
  }

  protected ModelImpl(String uri, Hashtable triples, LongTable lookup, boolean shared) {
    this.uri = uri;
    this.triples = triples;
    this.lookup = lookup;
    this.shared = shared;
  }

  /**
   * Set a base URI for the message.
   * Affects creating of new resources and serialization syntax.
   * <code>null</code> is just alright, meaning that
   * the serializer will use always rdf:about and never rdf:ID
   */
  public void setSourceURI(String uri) {
    this.uri = uri;
  }

  /**
   * Returns current base URI setting.
   */
  public String getSourceURI() {
    return uri;
  }

  // Model access

  /**
   * Number of triples in the model
   *
   * @return  number of triples
   */
  public int size() {
    return triples.size();
  }

  /**
   * Enumerate triples
   */
  public Enumeration elements() {
    return triples.elements();
  }

  /**
   * Tests if the model contains the given triple.
   *
   * @return  <code>true</code> if the triple belongs to the model;
   *          <code>false</code> otherwise.
   * @see     org.w3c.rdf.Statement
   */
  public synchronized boolean contains(Statement t) {
    return triples.contains(t);
  }

  // Model manipulation: add, remove, find

  /**
   * Adds a new triple to the model. A literal is created out of the string parameter <code>object</code>.
   * This method is just a shortcut.
   *
   * @see     org.w3c.rdf.Resource
   */
  public synchronized void add(Resource subject, Resource predicate, String object) {

    add(createStatement(subject, predicate, createLiteral(object)));
  }

  /**
   * Adds a new triple to the model. The object of the triple is an RDFNode.
   *
   * @see     org.w3c.rdf.Resource
   * @see     org.w3c.rdf.RDFNode
   */
  public synchronized void add(Resource subject, Resource predicate, RDFNode object) {

    add(createStatement(subject, predicate, object));
  }

  long getLookupValue(Resource subject, Resource predicate, RDFNode object) {

    int s1 = subject != null ? subject.hashCode() : 0;
    int s2 = predicate != null ? predicate.hashCode() : 0;
    int s3 = object != null ? object.hashCode() : 0;
    long l = ((long)s1 * 37 + s2) * 37 + s3;
    //    long l = (((((long)s1) << 16) ^ s2) << 16) ^ s3;
    //    System.err.println("lookup is " + l + " for " + subject + "," + predicate + "," + object);
    return l;
  }

  void addLookup(Statement t) {

    Resource s = t.subject();
    Resource p = t.predicate();
    RDFNode o = t.object();

    lookup.put(getLookupValue(null, null, o),    t);
    lookup.put(getLookupValue(null, p,    null), t);
    lookup.put(getLookupValue(null, p,    o),    t);
    lookup.put(getLookupValue(s,    null, null), t);
    lookup.put(getLookupValue(s,    null, o),    t);
    lookup.put(getLookupValue(s,    p,    null), t);
    lookup.put(getLookupValue(s,    p,    o),    t);
  }

  void removeLookup(Statement t) {

    Resource s = t.subject();
    Resource p = t.predicate();
    RDFNode o = t.object();

    lookup.remove(getLookupValue(null, null, o),    t);
    lookup.remove(getLookupValue(null, p,    null), t);
    lookup.remove(getLookupValue(null, p,    o),    t);
    lookup.remove(getLookupValue(s,    null, null), t);
    lookup.remove(getLookupValue(s,    null, o),    t);
    lookup.remove(getLookupValue(s,    p,    null), t);
    lookup.remove(getLookupValue(s,    p,    o),    t);
  }

  /**
   * Adds a new triple to the model.
   *
   * @see     org.w3c.rdf.Statement
   */
  public synchronized void add( Statement t ) {

    makeNonShared();

    triples.put(t, t);

    if(validLookup())
      addLookup(t);

    updateDigest(t);
  }

  synchronized void updateDigest(Statement t) {

    if(digest == null)
      return;

    Digest d = RDFDigestUtil.getStatementDigest(t);
    DigestUtil.xor(digest, d.getDigestBytes());
  }

  void makeNonShared() {

    if(shared) {
      triples = (Hashtable)triples.clone();
      lookup = (LongTable)lookup.clone();
      shared = false;
    }
  }

  boolean validLookup() {

    return lookup.size() > 0;
  }

  /**
   * Removes the triple from the model.
   *
   * @see     org.w3c.rdf.Statement
   */
  public synchronized void remove(Statement t) {

    makeNonShared();

    triples.remove(t);
    //    System.err.println("After " + triples.size());

    if(validLookup())
      removeLookup(t);

    updateDigest(t);
  }


  /**
   * General method to search for triples.
   * <code>null</code> input for any parameter will match anything.
   * <p>Example: <code>Model result = m.find( null, RDF.type, new Resource("http://...#MyClass") )</code>
   * <p>finds all instances of the class <code>MyClass</code>
   *
   * @see     org.w3c.rdf.Resource
   * @see     org.w3c.rdf.RDFNode
   */
  public synchronized Model find( Resource subject, Resource predicate, RDFNode object ) {

    // TODO: make everything synchronized by triples

    //    findCounter++;

    Model res = new ModelImpl();

    if(triples.size() == 0)
      return res;

    if(subject == null && predicate == null && object == null)
      return this.duplicate();

    boolean createLookup = !validLookup();

    //    if(createLookup)
    //      System.err.println("Creating lookup, size=" + triples.size() + ", shared=" + shared);

    Enumeration en = createLookup ? triples.elements() : lookup.multiget(getLookupValue(subject, predicate, object));

    while(en.hasMoreElements()) {

      Statement t = (Statement)en.nextElement();
      //      System.err.println("CHECKING: " + t);

      if(createLookup)
	addLookup(t);

      if(matchStatement(t, subject, predicate, object))
	res.add(t);
      //      counter++;
    }
    //    System.err.println("TRIPLES CHECKED: " + counter);
    return res;
  }


  boolean matchStatement(Statement triple, Resource subject, Resource predicate, RDFNode object) {

    if( subject != null && !triple.subject().equals(subject))
      return false;
      
    if(predicate != null && !triple.predicate().equals(predicate))
      return false;
      
    if(object != null && !triple.object().equals(object))
      return false;

    return true;
  }

  /**
   * Clone the model.
   */
  public Model duplicate() {

    //    return new ModelImpl(uri, (Hashtable)triples.clone(), lookup != null ? (LongTable)lookup.clone() : null, true);

    // creates a model that shares triples and lookup with this model
    return new ModelImpl(uri, triples, lookup, true);
  }

  public Object clone() {
    return duplicate();
  }

  /**
   * Creates empty model of the same Class
   */
  public Model create() {
    return new ModelImpl();
  }

  /**
   * Creates a resource out of a string
   */
  public Resource createResource(String str) {

    return new ResourceImpl(str);

  }

  /**
   * Creates a literal out of a string
   */
  public Literal createLiteral(String str) {

    return new LiteralImpl(str);

  }

  /**
   * Creates a triple
   */
  public Statement createStatement(Resource subject, Resource predicate, RDFNode object) {

    return new StatementImpl(subject, predicate, object);
  }


  // Resource implementation

  public String getLabel() {

    return getURI();
  }

  public Digest getDigest() {

    return this;
  }

  /*
  public synchronized Digest getDigest() {

    if(digest == null) {
      digest = RDFDigestUtil.getModelDigest(this);
    }
    return digest;
  }
  */
  public String getURI() {

    return "urn:rdf:model:" + DigestUtil.toHexString(getDigest());
  }

  // Digest implementation
  public String getDigestAlgorithm() {

    return Digest.MD5;
  }

  public synchronized byte[] getDigestBytes() {

    if(digest == null)
      digest = RDFDigestUtil.getModelDigest(this).getDigestBytes();

    byte[] result = new byte[16];
    System.arraycopy(digest, 0, result, 0, 16);
    return result;
  }

  /*
    Statement s1 = m.createStatement(m.createResource("uuid:timo"), m.createResource("is"), m.createResource("friend"));
    Statement s2 = m.createStatement(m.createResource("uuid:karsten"), m.createResource("is"), m.createResource("friend"));

    m.add(s1);
    System.err.println("With timo: " + m.getURI());

    m.add(s2);
    System.err.println("With timo & karsten: " + m.getURI());

    m.remove(s1);
    System.err.println("With karsten: " + m.getURI());

    m.remove(s2);
    System.err.println("Original: " + m.getURI());
    */

  /*
  protected void finalize() throws Throwable {
    System.err.println("Model " + this + " finalized: triples " + triples.size() + ", find ops: " + findCounter);
  }
  */

}
